import configparser
import datetime
import gzip
import json
import os
import shutil
import tempfile
import zipfile
# from logging import getLogger,  FileHandler, ERROR
from logging import ERROR, StreamHandler, getLogger
from pathlib import Path
from unittest import TestCase

import gridfs
from bson import DBRef, ObjectId
from bson.json_util import dumps
from pymongo import MongoClient, errors

from edman import DB, Config, Convert, File, Search
from edman.exceptions import EdmanDbProcessError


class TestFile(TestCase):
    db = None
    db_server_connect = False
    test_ini: dict = {}
    client = None

    @classmethod
    def setUpClass(cls):

        # 設定読み込み
        settings = configparser.ConfigParser()
        settings.read(Path.cwd() / 'ini' / 'test_db.ini')
        cls.test_ini = dict(settings.items('DB'))
        port = int(cls.test_ini['port'])
        cls.test_ini['port'] = port

        # DB作成のため、pymongoから接続
        cls.client = MongoClient(cls.test_ini['host'], cls.test_ini['port'])

        # 接続確認
        try:
            cls.client.admin.command('hello')
            cls.db_server_connect = True
            print('Use DB.')
        except errors.ConnectionFailure:
            print('Do not use DB.')

        if cls.db_server_connect:
            # adminで認証
            cls.client = MongoClient(
                username=cls.test_ini['admin_user'],
                password=cls.test_ini['admin_password'])
            # DB作成
            cls.client[cls.test_ini['db']].command(
                "createUser",
                cls.test_ini['user'],
                pwd=cls.test_ini['password'],
                roles=[
                    {
                        'role': 'dbOwner',
                        'db': cls.test_ini['db'],
                    },
                ],
            )
            # edmanのDB接続オブジェクト作成
            con = {
                'host': cls.test_ini['host'],
                'port': cls.test_ini['port'],
                'user': cls.test_ini['user'],
                'password': cls.test_ini['password'],
                'database': cls.test_ini['db'],
                'options': [f"authSource={cls.test_ini['db']}"]
            }
            cls.db = DB(con)
            cls.testdb = cls.db.get_db

            cls.logger = getLogger()

            # ログを画面に出力
            ch = StreamHandler()
            ch.setLevel(ERROR)  # ハンドラーにもそれぞれログレベル、フォーマットの設定が可能
            cls.logger.addHandler(ch)  # StreamHandlerの追加

            # ログをファイルに出力
            # fh = FileHandler('./tests.log')  # 引数には出力ファイルのパスを指定
            # fh.setLevel(
            #     ERROR)  # ハンドラーには、logger以下のログレベルを設定することは出来ない(この場合、DEBUGは不可)
            # cls.logger.addHandler(fh)  # FileHandlerの追加

    @classmethod
    def tearDownClass(cls):
        if cls.db_server_connect:
            # cls.clientはpymongo経由でDB削除
            # cls.testdb.dbはedman側の接続オブジェクト経由でユーザ(自分自身)の削除
            cls.client.drop_database(cls.test_ini['db'])
            cls.testdb.command("dropUser", cls.test_ini['user'])

    def setUp(self):

        if self.db_server_connect:
            self.file = File(self.db.get_db)
        else:
            self.file = File()
        self.config = Config()

    # def tearDown(self):
    #     pass

    @staticmethod
    def make_txt_files(dir_path, name='file_dl_list', suffix='.txt',
                       text='test', qty=1):
        # 添付ファイル用テキストファイル作成
        p = Path(dir_path)
        if qty == 1:
            filename = name + suffix
            save_path = p / filename
            with save_path.open('w') as f:
                f.write(text)
        else:
            for i in range(qty):
                filename = name + str(i) + suffix
                save_path = p / filename
                with save_path.open('w') as f:
                    f.write(text + str(i))
        return sorted(p.glob(name + '*' + suffix))

    def test_file_gen(self):

        # 正常系
        with tempfile.TemporaryDirectory() as tmp_dir:
            files = self.make_txt_files(tmp_dir, name='insert_file_ref_test',
                                        qty=2)
            expected = []
            for file_path in files:
                with file_path.open() as f:
                    expected.append((file_path.name, f.read()))

            # ファイル読み込み、テスト
            actual = []
            # files = tuple(sorted(p.glob('gen_test*.txt')))
            for idx, f in enumerate(self.file.file_gen(tuple(files))):
                filename, filedata = f
                actual.append((filename, filedata.decode()))
            self.assertListEqual(sorted(actual), sorted(expected))

    def test_file_list_attachment(self):

        # _ed_fileがなかった場合
        doc = {'name': 'NSX', 'ddd': 'aaa'}
        files_oid = [ObjectId(), ObjectId()]
        actual = self.file.file_list_attachment(doc, files_oid)
        expected = {'name': 'NSX', 'ddd': 'aaa', self.config.file: files_oid}
        self.assertDictEqual(expected, actual)

        # _ed_fileがすでに存在する場合
        oid1 = ObjectId()
        oid2 = ObjectId()

        files_oid = [oid1, oid2]
        doc = {'abc': '123', self.config.file: files_oid}
        at_files_oid = [ObjectId(), ObjectId()]
        actual = self.file.file_list_attachment(doc, at_files_oid)

        files_oid2 = [oid1, oid2]
        expected = {'abc': '123', self.config.file: files_oid2 + at_files_oid}

        self.assertDictEqual(actual, expected)

    def test_file_list_replace(self):
        # _ed_fileがある場合
        files_oid = [ObjectId(), ObjectId()]
        doc = {'name': 'NSX', self.config.file: files_oid}
        rep_files_oid = [ObjectId(), ObjectId()]
        actual = self.file.file_list_replace(doc, rep_files_oid)
        expected = {'name': 'NSX', self.config.file: rep_files_oid}
        self.assertDictEqual(expected, actual)

        # 空リストの場合
        doc = {'name': 'NSX', self.config.file: [ObjectId(), ObjectId()]}
        rep_files_oid = []
        actual = self.file.file_list_replace(doc, rep_files_oid)
        expected = {'name': 'NSX'}
        self.assertDictEqual(expected, actual)

        # _ed_fileがない場合(例外になる)
        doc = {'name': 'NSX'}
        rep_files_oid = [ObjectId(), ObjectId()]
        with self.assertRaises(ValueError):
            _ = self.file.file_list_replace(doc, rep_files_oid)

    def test_fs_delete(self):
        if not self.db_server_connect:
            return

        # 正常系
        with tempfile.TemporaryDirectory() as tmp_dir:
            fs_oids = []
            for i in tuple(
                    self.make_txt_files(tmp_dir, name='delete_test', qty=2)):
                with i.open('rb') as f:
                    self.fs = gridfs.GridFS(self.testdb)
                    fs_oids.append(
                        self.fs.put(f.read(), filename=str(i.name)))

            self.file.fs_delete(fs_oids)
            for i in fs_oids:
                with self.subTest(i=i):
                    self.assertFalse(self.fs.exists(i))

    def test_get_file_names(self):
        if not self.db_server_connect:
            return

        with tempfile.TemporaryDirectory() as tmp_dir:
            files = self.make_txt_files(tmp_dir, name='file_names', qty=2)

            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            for filename in files:
                with filename.open('rb') as f:
                    files_oid.append(
                        self.fs.put(f.read(), filename=filename.name))

            # DocをDBに入れる
            doc = {
                'test1': {
                    'test2': 'name',
                    'test3': {
                        'test4': [
                            {
                                'test5': 'moon',
                                self.config.file: files_oid,
                            },
                            {'test6': 'star'}
                        ]
                    }
                }
            }
            insert_result = self.testdb['structure_emb'].insert_one(doc)

            # 正常系テスト(embのみ.refとembの判断はget_file_ref()で行っているので割愛)
            oid = insert_result.inserted_id
            query = ['test1', 'test3', 'test4', '0']
            actual = self.file.get_file_names('structure_emb', oid, 'emb',
                                              query)
            expected = dict(zip(files_oid, [i.name for i in files]))
            self.assertDictEqual(actual, expected)

            # 正常系 ファイルがなかった場合 →空のリストを出力
            doc = {
                'test1': {
                    'test2': 'name2',
                    'test3': {
                        'test4': [
                            {'test5': 'moon2'},
                            {'test6': 'star2'}
                        ]
                    }
                }
            }
            insert_result = self.testdb['structure_ref'].insert_one(doc)
            oid = insert_result.inserted_id
            actual = self.file.get_file_names('structure_ref', oid, 'ref')
            self.assertDictEqual(actual, {})

            # 異常系
            # ドキュメントがなかった場合 →EdmanDbProcessErrorを補足
            oid = ObjectId()
            with self.assertRaises(EdmanDbProcessError):
                _ = self.file.get_file_names('structure_ref', oid, 'ref')

    def test__get_file_ref(self):
        # 正常系(ref)
        expected = [ObjectId(), ObjectId(), ObjectId()]
        doc = {
            'name': 'test1',
            self.config.file: expected,
            self.config.parent: DBRef('test0', ObjectId()),
            self.config.child: [DBRef('test2', ObjectId()),
                                DBRef('test3', ObjectId())]
        }
        structure = 'ref'
        actual = self.file.get_file_ref(doc, structure)
        self.assertListEqual(expected, actual)

        # 正常系(emb)
        expected = [ObjectId(), ObjectId(), ObjectId()]
        doc = {
            'test1': {
                'test2': 'name',
                'test3': {
                    'test4': [
                        {
                            'test5': 'moon',
                            self.config.file: expected,
                        },
                        {'test6': 'star'}
                    ]
                }
            }
        }
        structure = 'emb'
        query = ['test1', 'test3', 'test4', '0']
        actual = self.file.get_file_ref(doc, structure, query)
        self.assertListEqual(expected, actual)

    def test_delete(self):
        if not self.db_server_connect:
            return

        with tempfile.TemporaryDirectory() as tmp_dl_dir:

            # embの添付ファイル削除テスト
            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            dl_files = self.make_txt_files(tmp_dl_dir, name='file_delete')
            for filename in dl_files:
                with filename.open('rb') as f:
                    files_oid.append(
                        self.fs.put(f.read(), filename=filename.name))

            # DocをDBに入れる
            doc = {
                'test1': {
                    'test2': 'name',
                    'test3': {
                        'test4': [
                            {
                                'test5': 'moon',
                                self.config.file: files_oid,
                            },
                            {'test6': 'star'}
                        ]
                    }
                }
            }
            insert_result = self.testdb['structure_emb'].insert_one(doc)
            oid = str(insert_result.inserted_id)
            query = ['test1', 'test3', 'test4', '0']

            result = self.file.delete(files_oid[0], 'structure_emb', oid,
                                      'emb', query)
            self.assertTrue(result)
            # ファイルが消えたか検証
            self.assertFalse(self.fs.exists(files_oid[0]))

            # refの添付ファイル削除テスト
            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            for filename in dl_files:
                with filename.open('rb') as f:
                    files_oid.append(
                        self.fs.put(f.read(), filename=filename.name))

            # docをDBに入れる
            doc = {'name': 'test1',
                   self.config.file: files_oid,
                   self.config.parent: ObjectId(),
                   self.config.child: DBRef('child_col', ObjectId())}

            insert_result = self.testdb['structure_ref'].insert_one(doc)
            oid = str(insert_result.inserted_id)

            result = self.file.delete(files_oid[0], 'structure_ref', oid,
                                      'ref')
            self.assertTrue(result)
            # ファイルが消えたか検証
            self.assertFalse(self.fs.exists(files_oid[0]))

    def test__get_emb_files_list(self):

        # 正常系
        file_list = [ObjectId(), ObjectId()]

        doc = {
            "_id": ObjectId(),
            "position": "top",
            "username": "ryde",
            "bool": True,
            "structure_2": [
                {
                    "maker": "Ferrari",
                    "carname": "F355",
                    "power": 380,
                    "float_val": 4453.456
                },
                {
                    "maker": "HONDA",
                    "carname": "NSX",
                    "power": 280,
                    "float_val": 321.56,
                    "structure_3_1": [
                        {
                            "filename": "test1.txt",
                            "name": "添付ファイル1"
                        },
                        {
                            "filename": "test2.txt",
                            "name": "添付ファイル2"
                        }
                    ],
                    "structure_3_2": [
                        {
                            "filename": "test3.txt",
                            "name": "添付ファイル3",
                            "structure_4": {
                                "filename": "test4.txt",
                                "name": "添付ファイル4",
                                "structure_5": [
                                    {
                                        "url": "example2.com",
                                        "name": "テストURL2"
                                    },
                                    {
                                        "url": "example3.com",
                                        "name": "テストURL3"
                                    }
                                ]
                            },
                            "structure_6": [
                                {
                                    "url": "example_x.com",
                                    "name": "テストURL_x"
                                },
                                {
                                    "url": "example_y.com",
                                    "name": "テストURL_y",
                                    self.config.file: file_list,
                                }
                            ],
                            "structure_5": {
                                "url": "example.com",
                                "name": "テストURL1",
                                "structure_5": {
                                    "url": "example4.com",
                                    "name": "テストURL4"
                                }
                            }
                        }
                    ]
                }
            ]
        }

        query = ['structure_2', '1', 'structure_3_2', '0', 'structure_6', '1']
        actual = self.file._get_emb_files_list(doc, query)

        expected = file_list
        self.assertEqual(expected, actual)

    def test_zipped_json(self):

        # # 正常系
        s = {"document": [{"test1": "ABC"}, {"test2": 12345}]}
        raw_json = dumps(s, ensure_ascii=False, indent=4)
        encoded_json = raw_json.encode('utf-8')
        filename = 'jsonfile'

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir)
            ac_z_path = self.file.zipped_json(encoded_json, filename, p)

            # 圧縮前のjsonファイルを削除する
            first_file = filename + '.json'
            first = p / first_file
            first.unlink()

            # ac_z_pathを解凍して中身のファイル名がzipped_filepathと同じか調べる
            with zipfile.ZipFile(ac_z_path, 'r') as inputFile:
                inputFile.extractall(tmpdir)

            # 解凍されたディレクトリ名を取得する
            ps = list(p.glob('**/*.json'))[0]
            actual = ps.stem

        expected = filename
        self.assertEqual(expected, actual)

    def test_zipped_contents(self):
        if not self.db_server_connect:
            return

        with tempfile.TemporaryDirectory() as tmp_dl_dir:
            # 添付ファイル用テキストファイル作成

            test_var = 'test'
            name = 'file_dl_list.txt'
            input_filename = os.path.splitext(os.path.basename(name))[0]
            file_dl_list = self.make_txt_files(tmp_dl_dir, name=input_filename,
                                               text=test_var)

            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            for filename in file_dl_list:
                metadata = {'filename': filename.name, 'compress': None}
                with filename.open('rb') as f:
                    files_oid.append(self.fs.put(f.read(), **metadata))

            # docをDBに入れる
            parent_id = ObjectId()
            doc_id = ObjectId()
            child_id = ObjectId()
            parent_col = 'parent_col'
            doc_col = 'doc_col'
            child_col = 'child_col'
            insert_docs = [
                {
                    'col': parent_col,
                    'doc': {
                        '_id': parent_id,
                        'name': 'parent',
                        Config.child: [DBRef(doc_col, doc_id)]
                    },
                },
                {
                    'col': doc_col,
                    'doc': {
                        '_id': doc_id,
                        'name': 'doc',
                        Config.file: files_oid,
                        Config.parent: DBRef(parent_col, parent_id),
                        Config.child: [DBRef(child_col, child_id)]
                    }
                },
                {
                    'col': child_col,
                    'doc': {
                        '_id': child_id,
                        'name': 'child',
                        Config.parent: DBRef(doc_col, doc_id),
                    }
                }]
            result = {}
            for i in insert_docs:
                insert_result = self.testdb[i['col']].insert_one(i['doc'])
                result.update({i['col']: insert_result.inserted_id})

            # # ドキュメントを取得する
            search = Search(self.db)
            exclusion = ['_id', Config.file]
            docs = search.find(
                doc_col, {'_id': result[doc_col]},
                1, 1, exclusion)
            result_with_filepath, downloads = (
                self.file.get_fileref_and_generate_dl_list(
                    docs, "_ed_attachment"))
            res = search.generate_json_dict(result_with_filepath)

        # 実行
        raw_json = dumps(res, ensure_ascii=False, indent=4)
        encoded_json = raw_json.encode('utf-8')
        json_tree_file_name = 'json_tree'

        with tempfile.TemporaryDirectory() as tmp:
            p = Path(tmp)
            zip_filepath = self.file.zipped_contents(downloads,
                                                     json_tree_file_name,
                                                     encoded_json, p)
            # 解凍
            # print('unpack before',list(p.iterdir()))
            shutil.unpack_archive(zip_filepath, tmp)
            # print('unpack after', list(p.iterdir()))

            # 添付ファイルを読み込んでテキストの中身を抽出
            path_lists = [tmp, str(list(downloads.keys())[0]), name]
            with open(os.path.join(*path_lists)) as f:
                attached_s = f.read()
            path_lists = [tmp, json_tree_file_name + '.json']
            json_file_path = os.path.join(*path_lists)
            with open(json_file_path) as jf:
                j_data = json.load(jf)
            # 外に出すデータを組み立て(zip内部のjsonデータ、上記の添付ファイルデータ)
            actual_data = {'json_data': j_data, 'attached_data': attached_s}
        # テスト
        self.assertDictEqual(res, actual_data['json_data'])
        self.assertEqual(test_var, actual_data['attached_data'])


        # gzip圧縮されているデータの場合のテスト
        with tempfile.TemporaryDirectory() as tmp_dl_dir:
            # 添付ファイル用テキストファイル作成

            test_var = 'test'
            name = 'file_dl_list.txt'
            input_filename = os.path.splitext(os.path.basename(name))[0]
            file_dl_list = self.make_txt_files(tmp_dl_dir, name=input_filename,
                                               text=test_var)

            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            for filename in file_dl_list:
                metadata = {'filename': filename.name, 'compress': None}
                with filename.open('rb') as f:
                    # gzip圧縮
                    compressed = gzip.compress(f.read())
                    files_oid.append(self.fs.put(compressed, **metadata))

            # docをDBに入れる
            parent_id = ObjectId()
            doc_id = ObjectId()
            child_id = ObjectId()
            parent_col = 'parent_col'
            doc_col = 'doc_col'
            child_col = 'child_col'
            insert_docs = [
                {
                    'col': parent_col,
                    'doc': {
                        '_id': parent_id,
                        'name': 'parent',
                        Config.child: [DBRef(doc_col, doc_id)]
                    },
                },
                {
                    'col': doc_col,
                    'doc': {
                        '_id': doc_id,
                        'name': 'doc',
                        Config.file: files_oid,
                        Config.parent: DBRef(parent_col, parent_id),
                        Config.child: [DBRef(child_col, child_id)]
                    }
                },
                {
                    'col': child_col,
                    'doc': {
                        '_id': child_id,
                        'name': 'child',
                        Config.parent: DBRef(doc_col, doc_id),
                    }
                }]
            result = {}
            for i in insert_docs:
                insert_result = self.testdb[i['col']].insert_one(i['doc'])
                result.update({i['col']: insert_result.inserted_id})

            # # ドキュメントを取得する
            search = Search(self.db)
            exclusion = ['_id', Config.file]
            docs = search.find(
                doc_col, {'_id': result[doc_col]},
                1, 1, exclusion)
            result_with_filepath, downloads = (
                self.file.get_fileref_and_generate_dl_list(
                    docs, "_ed_attachment"))
            res = search.generate_json_dict(result_with_filepath)

        # 実行
        raw_json = dumps(res, ensure_ascii=False, indent=4)
        encoded_json = raw_json.encode('utf-8')
        json_tree_file_name = 'json_tree'

        with tempfile.TemporaryDirectory() as tmp:
            p = Path(tmp)
            zip_filepath = self.file.zipped_contents(downloads,
                                                     json_tree_file_name,
                                                     encoded_json, p)
            # 解凍
            # print('unpack before',list(p.iterdir()))
            shutil.unpack_archive(zip_filepath, tmp)
            # print('unpack after', list(p.iterdir()))

            # 添付ファイルを読み込んでテキストの中身を抽出
            path_lists = [tmp, str(list(downloads.keys())[0]), name]
            with open(os.path.join(*path_lists)) as f:
                attached_s = f.read()
            path_lists = [tmp, json_tree_file_name + '.json']
            json_file_path = os.path.join(*path_lists)
            with open(json_file_path) as jf:
                j_data = json.load(jf)
            # 外に出すデータを組み立て(zip内部のjsonデータ、上記の添付ファイルデータ)
            actual_data = {'json_data': j_data, 'attached_data': attached_s}
        # テスト
        self.assertDictEqual(res, actual_data['json_data'])
        self.assertEqual(test_var, actual_data['attached_data'])



    def test_get_fileref_and_generate_dl_list(self):

        if not self.db_server_connect:
            return

        # DBにdocsと添付ファイルを挿入する
        with tempfile.TemporaryDirectory() as tmp_dl_dir:
            name = 'file_dl_list.txt'
            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            input_filename = os.path.splitext(os.path.basename(name))[0]
            for filename in self.make_txt_files(tmp_dl_dir, input_filename):
                with filename.open('rb') as f:
                    files_oid.append(
                        self.fs.put(f.read(), filename=filename.name))
            # docをDBに入れる
            child_oid = ObjectId()
            doc = {'name': 'test1',
                   self.config.file: files_oid,
                   self.config.child: DBRef('child_col', child_oid)}
            insert_result = self.testdb[
                'test__get_fileref_and_generate_dl_list'].insert_one(doc)
            # ドキュメントを取得する
            res = self.testdb[
                'test__get_fileref_and_generate_dl_list'].find_one(
                {'_id': insert_result.inserted_id})
            # 実行
            new_docs, dl_list = self.file.get_fileref_and_generate_dl_list(
                res, '_ed_attachment')
            # 正常系 docsを比較
            expected = {'_id': insert_result.inserted_id, 'name': 'test1',
                        '_ed_attachment': [
                            str(insert_result.inserted_id) + '/' + name],
                        '_ed_child': DBRef('child_col', child_oid)}
            actual = new_docs
            self.assertDictEqual(expected, actual)

            # 正常系 DL処理用の辞書を比較
            expected = {insert_result.inserted_id: files_oid}
            actual = dl_list
            self.assertDictEqual(expected, actual)

    def test_generate_zip_filename(self):

        # 正常系
        # ファイル名を指定
        filename = 'testFileName'
        now = datetime.datetime.now()
        name = now.strftime('%Y%m%d%H%M%S')
        expected = name + filename + '.zip'
        actual = self.file.generate_zip_filename(filename)
        self.assertEqual(expected, actual)

        # ファイル名指定なし
        now = datetime.datetime.now()
        name = now.strftime('%Y%m%d%H%M%S')
        expected = name + '.zip'
        actual = self.file.generate_zip_filename()
        self.assertEqual(expected, actual)

        # 正常系
        # 文字列以外
        filename = 2545
        now = datetime.datetime.now()
        name = now.strftime('%Y%m%d%H%M%S')
        filename_str = str(filename)
        expected = name + filename_str + '.zip'
        actual = self.file.generate_zip_filename(filename)
        self.assertEqual(expected, actual)

    def test_grid_out(self):
        if not self.db_server_connect:
            return

        with tempfile.TemporaryDirectory() as tmp_dir:
            files = self.make_txt_files(tmp_dir, name='file_dl', qty=2)

            # ファイル読み込み、ファイルをgridfsに入れる
            files_oid = []
            self.fs = gridfs.GridFS(self.testdb)
            test_vars = {}
            for filename in files:
                with filename.open('rb') as f:
                    content = f.read()
                    # file_obj = gzip.compress(content, compresslevel=6)
                    # files_oid.append(
                    #     self.fs.put(file_obj, filename=filename.name,
                    #                 compress='gzip'))
                    files_oid.append(
                        self.fs.put(content, filename=filename.name))
                    test_vars.update({filename.name: content.decode()})

        with tempfile.TemporaryDirectory() as tmp_dl_dir:
            path = Path(tmp_dl_dir)

            grid_out_result = self.file._grid_out(files_oid, tmp_dl_dir)
            # 実行結果
            self.assertTrue(grid_out_result)

            # DB送信前のファイル名＆ファイルの中身と、ダウンロード後のファイル名&ファイルの中身を比較する
            expected = {}
            for dl_file in path.glob('*.txt'):
                with dl_file.open() as f:
                    expected.update({dl_file.name: f.read()})
            self.assertDictEqual(test_vars, expected)

    def test_upload(self):
        if not self.db_server_connect:
            return

        # refの場合
        # refデータ入力
        ref_json = {
            "structure_1": [
                {
                    "position": "top",
                    "username": "ryde",
                    "structure_2": [
                        {
                            "maker": "Ferrari",
                            "carname": "F355",
                            "power": 380,
                            "float_val": 4453.456
                        },
                        {
                            "maker": "HONDA",
                            "carname": "NSX",
                            "power": 280,
                            "float_val": 321.56,
                            "start_date": {"#date": "1981-04-23"},
                            "structure_3_1": [
                                {
                                    "filename": "test1.txt",
                                    "name": "添付ファイル1"
                                },
                                {
                                    "filename": "test2.txt",
                                    "name": "添付ファイル2"
                                }
                            ],
                            "structure_3_2": [
                                {
                                    "filename": "test3.txt",
                                    "name": "添付ファイル3",
                                    "structure_4": {
                                        "filename": "test4.txt",
                                        "name": "添付ファイル4",
                                        "structure_5": [
                                            {
                                                "url": "example2.com",
                                                "name": "テストURL2"
                                            },
                                            {
                                                "url": "example3.com",
                                                "name": "テストURL3"
                                            }
                                        ]
                                    },
                                    "structure_6": [
                                        {
                                            "url": "example_x.com",
                                            "name": "テストURL_x"
                                        },
                                        {
                                            "url": "example_y.com",
                                            "name": "テストURL_y"
                                        }
                                    ],
                                    "structure_5": {
                                        "url": "example.com",
                                        "name": "テストURL1",
                                        "structure_5": {
                                            "url": "example4.com",
                                            "name": "テストURL4"
                                        }
                                    }
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        convert = Convert()
        converted_edman = convert.dict_to_edman(ref_json, mode='ref')
        insert_result = self.db.insert(converted_edman)

        with tempfile.TemporaryDirectory() as tmp_dir:
            files = self.make_txt_files(tmp_dir, name='insert_file_ref_test',
                                        qty=2)
            expected = []
            for file_path in files:
                with file_path.open() as f:
                    expected.append((file_path.name, f.read()))

            # メソッド実行
            collection = 'structure_5'
            oid = [i[collection][0] for i in insert_result if collection in i][
                0]
            p_files = tuple([i for i in files])
            insert_file_result = self.file.upload(collection, oid, p_files,
                                                  'ref')

            # DBからデータ取得
            query = {'_id': oid}
            result = self.testdb[collection].find_one(query)

            # 取得したデータからgridfsのファイルを取得
            actual = []
            self.fs = gridfs.GridFS(self.testdb)
            for file_oid in result[self.config.file]:
                data = self.fs.get(file_oid)
                d = data.read()
                if hasattr(data, 'compress') and data.compress == 'gzip':
                    d = gzip.decompress(d)
                actual.append((data.filename, d.decode()))

            # テスト
            self.assertTrue(insert_file_result)
            self.assertListEqual(sorted(actual), sorted(expected))

        # embの場合
        data = {
            "position": "top",
            "structure_2": [
                {
                    "maker": "Ferrari",
                    "carname": "F355",
                    "power": 380,
                    "float_val": 4453.456
                },
                {
                    "maker": "HONDA",
                    "carname": "NSX",
                    "power": 280,
                    "float_val": 321.56,
                    "structure_3_1": [
                        {
                            "filename": "test1.txt",
                            "name": "添付ファイル1"
                        },
                        {
                            "filename": "test2.txt",
                            "name": "添付ファイル2"
                        }
                    ],
                    "structure_3_2": [
                        {
                            "filename": "test3.txt",
                            "name": "添付ファイル3",
                            "structure_4": {
                                "filename": "test4.txt",
                                "name": "添付ファイル4",
                                "structure_5": [
                                    {
                                        "url": "example2.com",
                                        "name": "テストURL2"
                                    },
                                    {
                                        "url": "example3.com",
                                        "name": "テストURL3"
                                    }
                                ]
                            },
                            "structure_6": [
                                {
                                    "url": "example_x.com",
                                    "name": "テストURL_x"
                                },
                                {
                                    "url": "example_y.com",
                                    "name": "テストURL_y"
                                }
                            ],
                            "structure_5": {
                                "url": "example.com",
                                "name": "テストURL1",
                                "structure_5": {
                                    "url": "example4.com",
                                    "name": "テストURL4"
                                }
                            }
                        }
                    ]
                }
            ]
        }
        insert_result = self.testdb['structure_emb'].insert_one(data)

        with tempfile.TemporaryDirectory() as tmp_dir:
            files_obj = self.make_txt_files(tmp_dir, name='emb_test', qty=2)
            expected = []
            for file_path in files_obj:
                with file_path.open() as f:
                    expected.append(f.read())

            collection = 'structure_emb'
            oid = insert_result.inserted_id
            query = ['structure_2', '1']

            files_obj2 = tuple([i for i in files_obj])
            # メソッド実行
            insert_file_emb_result = self.file.upload(
                collection, oid, files_obj2, 'emb', query)

            # ドキュメントをfindして出す
            result = self.testdb[collection].find_one({'_id': oid})

            # gridfsからデータ取得
            out_data = []
            self.fs = gridfs.GridFS(self.testdb)
            for oid in result['structure_2'][1][self.config.file]:
                data = self.fs.get(oid)
                f_data = data.read()
                if hasattr(data, 'compress') and data.compress == 'gzip':
                    f_data = gzip.decompress(f_data)

                out_data.append(f_data.decode())

            # DBデータとファイルのデータに差異はないか
            self.assertListEqual(sorted(expected), sorted(out_data))

            # メソッド成功時のフラグ
            self.assertTrue(insert_file_emb_result)

    def test_grid_in(self):
        if not self.db_server_connect:
            return

        # 正常系
        with tempfile.TemporaryDirectory() as tmp_dir:
            sample_files = self.make_txt_files(tmp_dir, name='grid_in_test',
                                               qty=2)

            td = tuple([p for p in sample_files])

            self.fs = gridfs.GridFS(self.testdb)
            actual = []
            for oid in self.file.grid_in(td):
                data = self.fs.get(oid)
                f_data = data.read().decode()
                actual.append([data.filename, f_data])

            expected = []
            for file_path in sample_files:
                with file_path.open() as f:
                    expected.append([file_path.name, f.read()])

            self.assertListEqual(sorted(actual), sorted(expected))

    def test_generate_file_path_dict(self):
        with tempfile.TemporaryDirectory() as tmp:
            tmp_p = Path(tmp)
            p = tmp_p / 'sub'
            p.mkdir()
            sample_files = self.make_txt_files(
                p,
                name='generate_file_path_dict_test',
                qty=2)
            file_list = list(map(str, sample_files))
            expected = {k: v for k, v in zip(file_list, sample_files)}
            actual = self.file.generate_file_path_dict(file_list, p)

            # paths = [(v, False) for k, v in actual.items()]
            # data = tuple(paths)
            # print('data', data)
            # grid_in_results = self.file.grid_in(data)
            # gf_inserted_dict = {}
            # for i, j in zip(actual, grid_in_results):
            #     gf_inserted_dict.update({i: j})
            # print('gf_inserted_dict', gf_inserted_dict)

            self.assertDictEqual(actual, expected)

    def test_generate_upload_list(self):
        json_dict = {
            "position": "top",
            "structure_2": [
                {
                    "maker": "Ferrari",
                    "carname": "F355",
                    "power": 380,
                    "_ed_attachment": ['aaa/01.jpg']
                },
                {
                    "maker": "Ferrari",
                    "carname": "458 Italia",
                    "_ed_attachment": ['bbb/01.jpg', 'bbb/02.jpg'],

                }],
            "structure_3": {
                "structure_4": {
                    "data1": 45,
                    "_ed_attachment": ['ccc/01.jpg']}
            },
            "structure_5": {
                "structure_6": {
                    "data1": ["123", "456"]
                },
            }

        }
        expected = ['aaa/01.jpg', 'bbb/01.jpg', 'bbb/02.jpg', 'ccc/01.jpg']
        actual = self.file.generate_upload_list(json_dict)
        # print(actual)
        self.assertListEqual(actual, expected)

    def test_json_rewrite(self):
        if not self.db_server_connect:
            return

        json_dict = {
            "position": "top",
            "structure_2": [
                {
                    "maker": "Ferrari",
                    "carname": "F355",
                    "power": 380,
                    "_ed_attachment": ['aaa/01.jpg']
                },
                {
                    "maker": "Ferrari",
                    "carname": "458 Italia",
                    "list_test": ["123", "456"],
                    "_ed_attachment": ['bbb/01.jpg', 'bbb/02.jpg'],

                }],
            "structure_3": {
                "structure_4": {
                    "data1": 45,
                    "_ed_attachment": ['ccc/01.jpg']}
            }
        }
        a01_file = ObjectId()
        b01_file = ObjectId()
        b02_file = ObjectId()
        c01_file = ObjectId()
        files_dict = {'aaa/01.jpg': a01_file,
                      'bbb/01.jpg': b01_file,
                      'bbb/02.jpg': b02_file,
                      'ccc/01.jpg': c01_file}
        expected = {
            "position": "top",
            "structure_2": [
                {
                    "maker": "Ferrari",
                    "carname": "F355",
                    "power": 380,
                    "_ed_file": [a01_file]
                },
                {
                    "maker": "Ferrari",
                    "carname": "458 Italia",
                    "list_test": ["123", "456"],
                    "_ed_file": [b01_file, b02_file]
                }],
            "structure_3": {
                "structure_4": {
                    "data1": 45,
                    "_ed_file": [c01_file]}
            }
        }
        actual = self.file.json_rewrite(json_dict, files_dict)
        # print(actual)
        self.assertDictEqual(actual, expected)



class Config:
    """
    このパッケージを利用したシステム上での共通の定義

    **デフォルトのままをおすすめします**

      | 定義を変更した場合、このパッケージを利用している他のシステムと、データ交換ができなくなります
      | DBにデータが入っている状態で、この定義を変更した場合、データが破壊される可能性があります

    | それでも変更したい場合は、単一、もしくは同じシステム内で定義を統一すること
    | その場合、他のシステムとデータ交換したくなった場合は独自に変換スクリプトを作成してください
    """
    # ドキュメント内でedmanが使用するリファレンス用のキー
    parent = '_ed_parent'  # 親のリファレンス情報
    child = '_ed_child'  # 子のリファレンス情報
    file = '_ed_file'  # Grid.fsのリファレンス情報

    # Grid.fsのデフォルトコレクション名
    fs_files = 'fs.files'  # ファイルコレクション名
    fs_chunks = 'fs.chunks'  # ファイルチャンクコレクション名

    # ユーザがJSON内で使用するキー
    # 日付に変換する場合
    # 例: "startDate": {"#date": "2020-07-01 00:00:00"}
    date = '#date'
    # JSON内で使用する添付ファイルディレクトリ用
    # 例: "_ed_attachment":["dir1/sample_photo.jpg", "dir1/experiment.cbf"]
    file_attachment = '_ed_attachment'



from datetime import datetime
# from logging import getLogger,  FileHandler, ERROR
from logging import ERROR, StreamHandler, getLogger
from unittest import TestCase

import dateutil.parser
from bson import ObjectId, errors

from edman import Config
from edman.utils import Utils


class TestUtils(TestCase):

    def setUp(self):
        self.config = Config()
        self.parent = self.config.parent
        self.child = self.config.child
        self.date = self.config.date
        self.file = self.config.file

    @classmethod
    def setUpClass(cls):
        cls.logger = getLogger()

        # ログを画面に出力
        ch = StreamHandler()
        ch.setLevel(ERROR)  # ハンドラーにもそれぞれログレベル、フォーマットの設定が可能
        cls.logger.addHandler(ch)  # StreamHandlerの追加

        # ログをファイルに出力
        # fh = FileHandler('./tests.log')  # 引数には出力ファイルのパスを指定
        # fh.setLevel(ERROR)  # ハンドラーには、logger以下のログレベルを設定することは出来ない(この場合、DEBUGは不可)
        # cls.logger.addHandler(fh)  # FileHandlerの追加


    def test__item_literal_check(self):
        # 正常系 リスト内が全てリテラルデータ
        data = [1, 2, 3]
        self.assertTrue(Utils.item_literal_check(data))

        # 正常系 リスト内にオブジェクトを含むリテラルデータ
        data = [1, 2, ObjectId()]
        self.assertTrue(Utils.item_literal_check(data))

        # 正常系 リスト内に辞書
        data = [1, 2, {'d': 'bb'}]
        self.assertFalse(Utils.item_literal_check(data))

        # 正常系 リスト内にリスト
        data = [1, 2, ['1', 2]]
        self.assertFalse(Utils.item_literal_check(data))

        # 正常系 入力が辞書
        data = {'d': '34'}
        self.assertFalse(Utils.item_literal_check(data))

    def test_doc_traverse(self):
        # 正常系1 対象のキーを削除する
        doc = {
            'a': '1',
            'b': [
                {
                    'd': {
                        'e': '4',
                        'f': [ObjectId(), ObjectId(), ObjectId()],
                        'g': '5'
                    }
                },
                {
                    'c': '3'
                }
            ]
        }
        del_keys = ['f']
        query = ['b', '0', 'd']

        def delete(d, keys):
            for key in keys:
                if key in d:
                    del d[key]

        actual = Utils.doc_traverse(doc, del_keys, query, delete)
        expected = {
            'a': '1',
            'b': [
                {
                    'd': {
                        'e': '4',
                        'g': '5'
                    }
                },
                {
                    'c': '3'
                }
            ]
        }
        self.assertDictEqual(actual, expected)

        # 例外のテスト1 クエリのインデックスの不備
        query = ['bbb', '2', 'eee', '2']
        doc = {
            'aaa': '123',
            'bbb': [
                {'ccc': '456'}, {'ddd': '789'},
                {'eee': [
                    {'fff': {'ans': 'OK'}}, {'ggg': '1', 'hhh': 'iii'}
                ]}
            ]
        }
        oids = [ObjectId(), ObjectId(), ObjectId()]
        with self.assertRaises(IndexError):
            _ = Utils.doc_traverse(doc, oids, query, delete)

        # 例外のテスト2 クエリの指定ミス
        query = ['bbb', '2', 'eee', 'xxx']
        doc = {
            'aaa': '123',
            'bbb': [
                {'ccc': '456'}, {'ddd': '789'},
                {'eee': [
                    {'fff': {'ans': 'OK'}}, {'ggg': '1', 'hhh': 'iii'}
                ]}
            ]
        }
        oids = [ObjectId(), ObjectId(), ObjectId()]
        with self.assertRaises(ValueError):
            _ = Utils.doc_traverse(doc, oids, query, delete)

    def test_conv_objectid(self):

        # 正常系 oidの場合
        oid = ObjectId()
        actual = Utils.conv_objectid(oid)
        self.assertIsInstance(actual, ObjectId)
        self.assertEqual(oid, actual)

        # 正常系 文字列の場合
        oid = ObjectId()
        actual = Utils.conv_objectid(str(oid))
        self.assertIsInstance(actual, ObjectId)
        self.assertEqual(oid, actual)

        # 異常系 oidにならない文字列
        oid = str(ObjectId())
        oid = oid[:-1]
        with self.assertRaises(errors.InvalidId):
            _ = Utils.conv_objectid(oid)

    def test__to_datetime(self):
        # datetime正常
        input_list = ['2018/11/20', '2018/11/20 13:48', '2018/01/01 00:00:00']
        for s in input_list:
            with self.subTest(s=s):
                actual = Utils.to_datetime(s)
                self.assertIsInstance(actual, datetime)

        # 入力値が文字列だがdatetime変換できない場合、または入力値が文字列以外
        input_list = [20181120, 201811201348, 20200101000000, '8月12日', None]
        for s in input_list:
            with self.subTest(s=s):
                actual = Utils.to_datetime(s)
                self.assertIsInstance(actual, str)

    def test__query_check(self):

        # 正常系
        query = ['bbb', '2', 'eee', '0', 'fff']
        doc = {
            'aaa': '123',
            'bbb': [
                {'ccc': '456'}, {'ddd': '789'},
                {'eee': [
                    {'fff': {'ans': 'OK'}}, {'ggg': '1'}
                ]}
            ]
        }
        actual = Utils.query_check(query, doc)
        self.assertIsInstance(actual, bool)
        self.assertTrue(actual)

        # 異常系 間違ったクエリ
        query = ['bbb', '2', 'eee', '1', 'fff']  # インデックスの指定ミスを想定
        doc = {
            'aaa': '123',
            'bbb': [
                {'ccc': '456'}, {'ddd': '789'},
                {'eee': [
                    {'fff': {'ans': 'OK'}}, {'ggg': '1'}
                ]}
            ]
        }
        actual = Utils.query_check(query, doc)
        self.assertIsInstance(actual, bool)
        self.assertFalse(actual)

    def test_item_delete(self):
        # 正常系
        doc = {
            self.parent: ObjectId(),
            self.child: [ObjectId(), ObjectId()],
            self.file: [ObjectId(), ObjectId()],
            'param': 'OK'
        }
        actual = Utils.item_delete(
            doc, ('_id', self.parent, self.child, self.file))
        expected = {'param': 'OK'}
        self.assertDictEqual(actual, expected)

    def test_child_combine(self):
        # データ構造のテスト
        test_data = [
            [
                {'collection_A': {'name': 'NSX'}},
                {'collection_A': {'name': 'F355'}},
                {'collection_B': {'power': 280}}
            ]
        ]
        actual = [i for i in Utils.child_combine(test_data)]
        self.assertIsInstance(actual, list)
        self.assertEqual(2, len(actual[0]['collection_A']))

    def test_field_name_check(self):

        illegals = [None, '', '$aa', '.aa']
        for i in illegals:
            with self.subTest(i=i):
                actual = Utils.field_name_check(i)
                self.assertFalse(actual)

        # 文字列以外の方は文字列に変換される
        actual = Utils.field_name_check(455)
        self.assertTrue(actual)

    def test_collection_name_check(self):

        illegals = [None, '', '$aaa', 'aaa$b', 'system.aaa', '#aaa', '@aaa']
        for i in illegals:
            with self.subTest(i=i):
                actual = Utils.collection_name_check(i)
                self.assertFalse(actual)

        # 文字列以外の方は文字列に変換される
        actual = Utils.collection_name_check(345)
        self.assertTrue(actual)

    def test_type_cast_conv(self):
        # 正常系 変換テスト
        input_l = ['str', 'int', 'float', 'bool', 'datetime']
        expected = [str, int, float, bool, dateutil.parser.parse]
        actual = [Utils.type_cast_conv(i) for i in input_l]
        self.assertEqual(expected, actual)

        # 正常系 str変更テスト
        input_l = ['str', '12']
        expected = [str, str]
        actual = [Utils.type_cast_conv(i) for i in input_l]
        self.assertEqual(expected, actual)



from datetime import datetime
from logging import INFO, getLogger

from bson import DBRef, ObjectId
from bson import errors as bson_errors
from pymongo import errors

from edman import Config
from edman.exceptions import (EdmanDbProcessError, EdmanFormatError,
                              EdmanInternalError)
from edman.utils import Utils


# from collections import deque

class Search:
    """
    検索関連クラス

    """

    def __init__(self, db=None) -> None:
        config = Config()  # システム環境用の設定を読み込む
        self.parent = config.parent
        self.child = config.child
        self.date = config.date
        self.file = config.file
        self.db = db

        # ログ設定(トップに伝搬し、利用側でログとして取得してもらう)
        self.logger = getLogger(__name__)
        self.logger.setLevel(INFO)
        self.logger.propagate = True

        if self.db is not None:
            self.connected_db = db.get_db

    def find(self, collection: str, query: dict, parent_depth=0,
             child_depth=0, exclusion=None) -> dict:
        """
        検索用メソッド

        :param str collection: 対象コレクション
        :param dict query: 検索クエリ
        :param int parent_depth: 親の指定深度
        :param int child_depth: 子の指定深度
        :param None or list exclusion:除外するリファレンスキー 例 ['_ed_file']
        :return: result 親 + 自分 + 子の階層構造となった辞書データ
        :rtype: dict
        """

        coll_filter = {"name": {"$regex": r"^(?!system\.)"}}
        if collection not in self.connected_db.list_collection_names(
                filter=coll_filter):
            raise EdmanDbProcessError('コレクションが存在しません')

        query = self._objectid_replacement(query)
        self_result = self._get_self(query, collection)
        if self_result is None:
            raise EdmanDbProcessError('データを取得できませんでした')
        reference_point_result = self.db.get_reference_point(
            self_result[collection])

        parent_result = None
        if reference_point_result[self.parent]:
            parent_result = self._get_parent(self_result, parent_depth)

        children_result = None
        if reference_point_result[self.child]:
            children_result = self.db.get_child(self_result, child_depth)

        # 親も子も存在しない時はselfのみ
        result = self_result

        # 子データが存在する時だけselfとマージ
        if children_result:
            self_key = list(self_result.keys())[0]
            self_result[self_key].update(children_result)

        # 親データが存在する時だけselfとマージ
        if parent_result:
            result = self._merge_parent(parent_result, result)

        # JSONデータ用に変換
        result = self.generate_json_dict(result, include=exclusion)

        return result

    def _merge_parent(self, parent_result: dict, self_result: dict) -> dict:
        """
        親データに家族データをマージする

        :param dict parent_result:マージ前の親データ
        :param dict self_result:マージ済み家族データ
        :return: parent_result 親と自分と子をマージしたデータ
        :rtype: dict
        """

        def recursive(data: dict):
            """
            親データ内で一番深いドキュメントに家族データを挿入する

            :param dict data:
            :return:
            """
            for key in data.keys():
                # 親が複数存在することはありえないので辞書のみで可
                if isinstance(data[key], dict):
                    if '_id' in data[key]:
                        if parent_id == data[key]['_id']:
                            data[key].update(self_result)
                    recursive(data[key])

        self_doc = list(self_result.values())[0]
        parent_id = self_doc[self.parent].id
        recursive(parent_result)  # parent_resultにself_resultを入れる
        return parent_result

    @staticmethod
    def _objectid_replacement(query: dict) -> dict:
        """
        ObjectIdのチェックと変換

        :param dict query:
        :return: query
        :rtype: dict
        """
        if '_id' in query:
            try:
                query['_id'] = ObjectId(query['_id'])
            except bson_errors.InvalidId:
                raise
        return query

    def _get_self(self, query: dict, collection: str) -> dict | None:
        """
        自分自身のドキュメント取得

        :param dict query:
        :param str collection:
        :return:
        :rtype: dict or None
        """
        result = None
        try:
            docs = list(self.connected_db[collection].find(query))
        except errors.OperationFailure:
            raise EdmanDbProcessError('ドキュメントが取得できませんでした')
        else:
            if len(docs):
                result = {collection: docs[0]}
        return result

    def _get_parent(self, self_doc: dict, depth: int) -> dict | None:
        """
        | 親となるドキュメントを取得
        | depthで深度を設定し、階層分取得する

        :param dict self_doc:
        :param int depth:
        :return: result
        :rtype: dict or None
        """

        def recursive(doc):
            """
            再帰でドキュメントを取得
            DBReferenceを利用し、設定されている深度を減らしながら再帰
            """
            if self.parent in doc:
                parent = self.connected_db.dereference(doc[self.parent])
                parent_collection = doc[self.parent].collection
                data.append({parent_collection: parent})
                nonlocal depth
                depth -= 1
                if depth > 0:
                    recursive(parent)

        if depth > 0:
            data: list = []  # recによって書き換えられる
            recursive(list(self_doc.values())[0])
            result = self._build_to_doc_parent(data)
        else:
            result = None
        return result

    @staticmethod
    def _build_to_doc_parent(parent_data_list: list) -> dict | None:
        """
        | 親の検索結果（リスト）を入れ子辞書に組み立てる
        |
        | 入力値であるparent_data_listについて、
        | parentに近い方から順番に並んでいる(一番最後がrootまたはrootに近い方)

        :param list parent_data_list:
        :return: result
        :rtype: dict or None
        """
        result = None
        for read_data in parent_data_list:
            if result is not None:
                buff = read_data
                collection = list(buff.keys())[0]
                doc = list(buff.values())[0]
                doc.update(result)
                result = {collection: doc}
            else:
                result = read_data
        return result

    def process_data_derived_from_mongodb(self, result_dict: dict,
                                          exclusion=None) -> dict:
        """
        MongoDB依存の項目を処理する::
          _idやrefの削除
          型をJSONに合わせる

          廃止予定 代替 generate_json_dict()

        :param dict result_dict:
        :param List or None exclusion:
            default ['_id', self.parent, self.child, self.file]
        :return: result_dict
        :rtype: dict
        """
        result_dict = self.generate_json_dict(result_dict, include=exclusion)

        return result_dict

    def _format_datetime(self, item: datetime) -> dict[str, str]:
        """
        datetime型なら書式変更して辞書に入れる

        :param item:
        :type item: datetime
        :return: result
        :rtype: dict
        """
        return {self.date: item.strftime("%Y-%m-%d %H:%M:%S")}

    def doc2(self, collection: str, oid: ObjectId | str,
             exclude_keys=None) -> dict:
        """
        指定するドキュメントを取得する
        doc()の置き換え版 リファクタリング完了後にdoc()を削除する
        embは対象外

        :param str collection:
        :param ObjectId oid:
        :param None or list exclude_keys: e.g.
            ['_id', 'parent', 'child', 'file']
        :return: result
        :rtype: dict
        """

        if isinstance(exclude_keys, list):
            exclude_keys = tuple(exclude_keys)
        elif exclude_keys is not None:
            raise EdmanFormatError('exclude_keysはlistで指定してください')

        doc = self.connected_db[collection].find_one(
            {'_id': Utils.conv_objectid(oid)})
        if doc is None:
            result = {}
        else:
            result = Utils.item_delete(dict(doc),
                                       exclude_keys) if exclude_keys else doc

        return result

    def get_tree(self, collection: str, oid: ObjectId, include=None) -> dict:
        """
        oidで指定するドキュメントが所属するツリーを全て取得する

        :param str collection:
        :param ObjectId oid:
        :param None or list include: e.g. ['_id', 'parent', 'child', 'file']
        :return: result
        :rtype: dict
        """

        self_doc = self.doc2(collection, oid)
        root_ref = self.db.get_root_dbref(self_doc)

        # root_refがNoneの場合は親ドキュメント
        if root_ref is None:
            root_doc = self_doc
            root_ref = DBRef(collection, oid)
        else:
            root_doc = self.doc2(root_ref.collection, root_ref.id)

        children = self.db.get_child_all({root_ref.collection: root_doc})

        parents = []
        for d in list(children.values()):
            for i in d:
                parents.append(i[self.parent])

        if all([i for i in parents if i == root_ref]):
            result_docs = dict(**root_doc, **children)
            tree = {root_ref.collection: result_docs}
            result = self.generate_json_dict(tree, include=include)

        else:
            raise EdmanInternalError(
                'ルートのドキュメントと子要素が一致しません'
                + root_ref.collection + ':' + str(root_ref.id))

        return result

    def generate_json_dict(self, result_dict: dict, include=None) -> dict:
        """
        edman依存の項目を処理する::
          _idとrefの削除
          型をJSONに合わせる

        process_data_derived_from_mongodb()の置き換え版
        process_data_derived_from_mongodb()は廃止予定

        :param dict result_dict:
        :param List or None include:
            e.g. ['_id', self.parent, self.child, self.file]
        :return: result_dict
        :rtype: dict
        """
        default_refs = ['_id', self.parent, self.child, self.file]
        if include is None:
            refs = tuple(default_refs)
        else:
            if not isinstance(include, list):
                raise ValueError('listもしくはNoneが必要です')
            else:
                for i in include:
                    if i not in default_refs:
                        raise ValueError(
                            f"{default_refs}の中から選択する必要があります")
            # デフォルトの値からexclusionを差し引く
            refs = tuple(set(default_refs) - set(include))

        def recursive(data: dict):
            # idとrefの削除
            for key, val in data.items():
                if isinstance(data[key], dict):
                    recursive(Utils.item_delete(data[key], refs))
                # リストデータは中身を型変換する
                elif isinstance(data[key], list) and Utils.item_literal_check(
                        data[key]):
                    data[key] = [self._format_datetime(j)
                                 if isinstance(j, datetime) else j
                                 for j in data[key]]
                elif isinstance(data[key], list):
                    for item in data[key]:
                        recursive(Utils.item_delete(item, refs))
                else:
                    try:  # 型変換
                        if isinstance(data[key], datetime):
                            data[key] = self._format_datetime(data[key])
                    except Exception:
                        raise

        recursive(result_dict)
        return result_dict

    # def logger_test(self):
    #     self.logger.error('logger test メソッド内 エラー')
    #     raise EdmanInternalError('logger test メソッド内 例外')

    # def get_ref_depth_bfs(self, collection, oid):
    #     # 子要素の最大の深さを取得する
    #     max_depth = 0
    #     doc = self.doc2(collection, oid)
    #     if self.child not in doc:
    #         raise EdmanDbProcessError('子要素が存在しません')  # これではなく0を返すべき?
    #
    #     q = deque([])  # キュー
    #     q.append(doc[self.child])
    #
    #     while len(q) > 0:
    #
    #         children = q.popleft()  # キュー取りだし（先頭）
    #         # このへんに別の配列を用意?
    #         for ref in children: # whileにする?
    #             d = self.connected_db.dereference(ref)
    #             if self.child in d:
    #                 q.append(d[self.child]) #ここでキューに言えるのは変?
    #
    #         max_depth += 1  # これを増やすタイミングは配列が空になったとき?
    #
    #     return max_depth



import configparser
import copy
from datetime import datetime
# from logging import getLogger,  FileHandler, ERROR
from logging import ERROR, StreamHandler, getLogger
from pathlib import Path
from unittest import TestCase

import dateutil.parser
from bson import DBRef, ObjectId, errors
from pymongo import MongoClient
from pymongo import errors as py_errors

from edman import DB, Config, Convert, Search


class TestSearch(TestCase):
    db_server_connect = False
    test_ini: dict = {}
    client = None
    testdb = None

    @classmethod
    def setUpClass(cls):
        # 設定読み込み
        settings = configparser.ConfigParser()
        settings.read(Path.cwd() / 'ini' / 'test_db.ini')
        cls.test_ini = dict(settings.items('DB'))
        port = int(cls.test_ini['port'])
        cls.test_ini['port'] = port

        # DB作成のため、pymongoから接続
        cls.client = MongoClient(cls.test_ini['host'], cls.test_ini['port'])

        # 接続確認
        try:
            cls.client.admin.command('hello')
            cls.db_server_connect = True
            print('Use DB.')
        except py_errors.ConnectionFailure:
            print('Do not use DB.')

        if cls.db_server_connect:
            # adminで認証
            cls.client = MongoClient(
                username=cls.test_ini['admin_user'],
                password=cls.test_ini['admin_password'])
            # DB作成
            cls.client[cls.test_ini['db']].command(
                "createUser",
                cls.test_ini['user'],
                pwd=cls.test_ini['password'],
                roles=[
                    {
                        'role': 'dbOwner',
                        'db': cls.test_ini['db'],
                    },
                ],
            )
            # edmanのDB接続オブジェクト作成
            con = {
                'host': cls.test_ini['host'],
                'port': cls.test_ini['port'],
                'user': cls.test_ini['user'],
                'password': cls.test_ini['password'],
                'database': cls.test_ini['db'],
                'options': [f"authSource={cls.test_ini['db']}"]
            }
            db = DB(con)
            cls.testdb = db.get_db
            cls.search = Search(db)
            cls.db = db
        else:
            cls.search = Search()

        cls.logger = getLogger()

        # ログを画面に出力
        ch = StreamHandler()
        ch.setLevel(ERROR)  # ハンドラーにもそれぞれログレベル、フォーマットの設定が可能
        cls.logger.addHandler(ch)  # StreamHandlerの追加

        # ログをファイルに出力
        # fh = FileHandler('./tests.log')  # 引数には出力ファイルのパスを指定
        # fh.setLevel(ERROR)  # ハンドラーには、logger以下のログレベルを設定することは出来ない(この場合、DEBUGは不可)
        # cls.logger.addHandler(fh)  # FileHandlerの追加


    @classmethod
    def tearDownClass(cls):
        if cls.db_server_connect:
            # cls.clientはpymongo経由でDB削除
            # cls.testdb.dbはedman側の接続オブジェクト経由でユーザ(自分自身)の削除
            cls.client.drop_database(cls.test_ini['db'])
            cls.testdb.command("dropUser", cls.test_ini['user'])

    def setUp(self):
        self.config = Config()
        self.parent = self.config.parent
        self.child = self.config.child
        self.date = self.config.date
        self.file = self.config.file

    def tearDown(self):
        if self.db_server_connect:
            # システムログ以外のコレクションを削除
            collections_all = self.testdb.list_collection_names()
            log_coll = 'system.profile'
            if log_coll in collections_all:
                collections_all.remove(log_coll)
            for collection in collections_all:
                self.testdb.drop_collection(collection)

    def test__merge_parent(self):

        # データ構造のテスト
        # 本来は親子関係なのでchild項目も存在するが、テストに関係ないので省略
        dummy_id = ObjectId()
        dummy_id2 = ObjectId()
        parent_id = ObjectId()
        parent_data = {
            'parent_col': {
                '_id': dummy_id,
                'a': 'aa',
                'parent_col2': {
                    '_id': dummy_id2,
                    'b': 'bb',
                    self.parent: DBRef('parent_col', dummy_id),
                    'parent_col3': {
                        '_id': parent_id,
                        'c': 'cc',
                        self.parent: DBRef('parent_col2', dummy_id2)
                    }
                }
            }
        }
        self_data = {
            'self_col': {
                'car_name': 'F355',
                self.parent: DBRef('parent_col3', parent_id)
            }
        }
        actual = self.search._merge_parent(parent_data, self_data)
        expected = \
            actual['parent_col']['parent_col2']['parent_col3']['self_col'][
                'car_name']

        self.assertIsInstance(actual, dict)
        self.assertEqual('F355', expected)

    def test__objectid_replacement(self):
        # 正常系
        query = {'_id': '5bf4f3ce9663380fc50d6dbd'}
        actual = self.search._objectid_replacement(query)
        self.assertIsInstance(actual['_id'], ObjectId)

        # 異常系
        query = {'_id': 'dragon'}
        with self.assertRaises(errors.InvalidId):
            _ = self.search._objectid_replacement(query)

    def test__get_self(self):
        if not self.db_server_connect:
            return

        # テストデータをDBに挿入
        data = {'test_data': 'test'}
        test_collection = 'get_self_test'
        db = self.client[self.test_ini['db']]
        insert_result = db[test_collection].insert_one(data)

        # メソッドを実行してデータを取ってくる dictや構造
        # 作成したデータと取得したデータの差異
        query = {'_id': insert_result.inserted_id}
        actual = self.search._get_self(query, test_collection)
        self.assertIsInstance(actual, dict)
        self.assertEqual(sorted(list(data.keys())),
                         sorted(list(actual[test_collection].keys())))

        # 存在しないデータの時はNoneを返す
        query = {'_id': insert_result.inserted_id, 'test_data': 'foo'}
        actual = self.search._get_self(query, test_collection)
        self.assertIsNone(actual)

    def test__get_parent(self):
        if not self.db_server_connect:
            return

        # テストデータをDBに挿入
        db = self.client[self.test_ini['db']]
        data1_id = ObjectId()
        data2_id = ObjectId()
        data3_id = ObjectId()
        parent_coll = 'parent_coll'
        parent2_coll = 'parent2_coll'
        self_coll = 'self_coll'
        data1 = {
            '_id': data1_id,
            'data1': 'test',
            self.child: [DBRef(parent2_coll, data2_id)]
        }
        _ = db[parent_coll].insert_one(data1)
        data2 = {
            '_id': data2_id,
            'data2': 'test',
            self.parent: DBRef(parent_coll, data1_id),
            self.child: [DBRef(self_coll, data3_id)]
        }
        _ = db[parent2_coll].insert_one(data2)
        data3 = {
            '_id': data3_id,
            'data3': 'test',
            self.parent: DBRef(parent2_coll, data2_id)
        }
        _ = db[self_coll].insert_one(data3)

        self_result = {self_coll: data3}
        actual = self.search._get_parent(self_result, depth=2)
        self.assertIsInstance(actual, dict)

        for k, v in actual.items():
            # 構造のチェック
            self.assertEqual(k, parent_coll)
            self.assertEqual(v['_id'], data1['_id'])
            self.assertIsInstance(v[parent2_coll], dict)
            self.assertEqual(v[parent2_coll]['_id'], data2_id)

            # 親データ内のネストチェック
            self.assertEqual(v[self.child][0].id, v[parent2_coll]['_id'])
            self.assertEqual(v['_id'], v[parent2_coll][self.parent].id)

            # selfとparentのネストチェック
            self.assertEqual(v[parent2_coll][self.child][0].id, data3['_id'])
            self.assertEqual(v[parent2_coll]['_id'], data3[self.parent].id)


        # テスト
        # parent_col = 'Beamtime'
        # target_col = 'expInfo'
        # d = {
        #     parent_col:
        #         {
        #             "test_data": "test",
        #             target_col: [
        #                 {
        #                     "layer_test1b": "data",
        #                     "layer_test1a": [
        #                         {
        #                             "test2_data": "data",
        #                             "layer_test2": [
        #                                 {
        #                                     "layer_test3a": "data",
        #                                     "layer_test3b": [
        #                                         {
        #                                             "layer_test4a": "data",
        #                                             "layer_test4b": [
        #                                                 {
        #                                                     "layer_test5a": "data"
        #                                                 }
        #                                             ]
        #                                         }
        #                                     ]
        #                                 }
        #                             ]
        #                         },
        #                         {
        #                             "test3_data": "data",
        #                             "layer_test2": [
        #                                 {
        #                                     "layer_test3a_2": "data3_2",
        #                                     "layer_test3b": [
        #                                         {
        #                                             "layer_test4a_2": "data_2",
        #                                             "layer_test4b": [
        #                                                 {
        #                                                     "layer_test5a_2": "data_2"
        #                                                 },
        #                                                 {
        #                                                     "elen": "elen",
        #                                                     "layer_test5": [
        #                                                         {
        #                                                             "mario": "mario"
        #                                                         },
        #                                                         {
        #                                                             "mario2": "mario22"
        #                                                         },
        #                                                     ]
        #                                                 }
        #                                             ]
        #                                         }
        #                                     ]
        #                                 }
        #                             ]
        #                         }
        #                     ]
        #                 }
        #             ]
        #         }
        # }
        #
        # convert = Convert()
        # insert_result = self.db.insert(convert.dict_to_edman(d))
        #
        # # インサート結果からtopのoidを取得
        # b = []
        # for i in insert_result:
        #     for k, v in i.items():
        #         if k == parent_col:
        #             b.extend(v)
        # insert_root_oid = b[0]
        #
        # # rootのドキュメントを取得
        # docs = self.search.doc2(parent_col, insert_root_oid)
        #
        # # 中間子要素と最後の子要素のテスト用データ取得のために子要素を全部取得
        # o = self.db.get_child_all({parent_col: docs})
        # # x = self.search.generate_json_dict(o, include=['_id', Config.parent,
        # #                                                Config.child])
        # x = self.search.generate_json_dict(o, include=[Config.parent,
        #                                                Config.child])
        #
        # # 最下層の子要素を取得
        # # self_doc_id = (x['expInfo'][0]['layer_test1a'][1]['layer_test2'][0]
        # # ['layer_test3b'][0]['layer_test4b'][1]['layer_test5'][0]['_id'])
        #
        # lower_doc = (x['expInfo'][0]['layer_test1a'][1]['layer_test2'][0]
        # ['layer_test3b'][0]['layer_test4b'][1]['layer_test5'][0])
        # print('lower_doc', lower_doc)
        # a = self.search._get_parent({'layer_test5':lower_doc}, depth=4)
        # print('a', a)

    # def test_get_ref_depth_bfs(self):
    #     if not self.db_server_connect:
    #         return
    #
    #     parent_col = 'Beamtime'
    #     target_col = 'expInfo'
    #     d = {
    #         parent_col:
    #             {
    #                 "test_data": "test",
    #                 target_col: [
    #                     {
    #                         "layer_test1b": "data",
    #                         "layer_test1a": [
    #                             {
    #                                 "test2_data": "data",
    #                                 "layer_test2": [
    #                                     {
    #                                         "layer_test3a": "data",
    #                                         "layer_test3b": [
    #                                             {
    #                                                 "layer_test4a": "data",
    #                                                 "layer_test4b": [
    #                                                     {
    #                                                         "layer_test5a": "data"
    #                                                     }
    #                                                 ]
    #                                             }
    #                                         ]
    #                                     }
    #                                 ]
    #                             },
    #                             {
    #                                 "test3_data": "data",
    #                                 "layer_test2": [
    #                                     {
    #                                         "layer_test3a_2": "data3_2",
    #                                         "layer_test3b": [
    #                                             {
    #                                                 "layer_test4a_2": "data_2",
    #                                                 "layer_test4b": [
    #                                                     {
    #                                                         "layer_test5a_2": "data_2"
    #                                                     },
    #                                                     {
    #                                                         "elen": "elen",
    #                                                         "layer_test5": [
    #                                                             {
    #                                                                 "mario": "mario"
    #                                                             },
    #                                                             {
    #                                                                 "mario2": "mario22"
    #                                                             },
    #                                                         ]
    #                                                     }
    #                                                 ]
    #                                             }
    #                                         ]
    #                                     }
    #                                 ]
    #                             }
    #                         ]
    #                     }
    #                 ]
    #             }
    #     }
    #
    #     convert = Convert()
    #     insert_result = self.db.insert(convert.dict_to_edman(d))
    #
    #     # インサート結果からtopのoidを取得
    #     b = []
    #     for i in insert_result:
    #         for k, v in i.items():
    #             if k == parent_col:
    #                 b.extend(v)
    #     insert_root_oid = b[0]
    #
    #     # rootのドキュメントを取得
    #     docs = self.search.doc2(parent_col, insert_root_oid)
    #
    #     # 中間子要素と最後の子要素のテスト用データ取得のために子要素を全部取得
    #     o = self.db.get_child_all({parent_col: docs})
    #     # x = self.search.generate_json_dict(o, include=['_id', Config.parent,
    #     #                                                Config.child])
    #     x = self.search.generate_json_dict(o, include=['_id', Config.parent,
    #                                                    Config.child])
    #     target_doc = x['expInfo'][0]
    #     print('target_doc', target_doc)
    #
    #     result = self.search.get_ref_depth_bfs('expInfo',target_doc['_id'])
    #     print(result)


    def test__build_to_doc_parent(self):
        # データ構造のテスト
        # parentに近い方から順番に並んでいる(一番最後がrootまたはrootに近い方)
        parent_data_list = [
            {
                'parent_2': {
                    'car_name': 'STORATOS'
                }
            },
            {
                'parent_1': {
                    'program': 'python'
                }
            }
        ]
        actual = self.search._build_to_doc_parent(parent_data_list)
        self.assertIsInstance(actual, dict)
        expected = actual['parent_1']['parent_2']['car_name']
        self.assertEqual('STORATOS', expected)

    def test_generate_json_dict(self):

        # 正常系
        data = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }

        def _item_literal_check(list_child):
            result = True
            if isinstance(list_child, dict):
                result = False
            if isinstance(list_child, list):
                for j in list_child:
                    if isinstance(j, dict) or isinstance(j, list):
                        result = False
                        break
            return result

        # リファレンス削除チェック用メソッド
        def detect_ref(d):
            if self.parent == d or self.child == d:
                raise ValueError('parent or child, was not deleted')

        # 日付変換チェック用メソッド
        def detect_date(d):
            if isinstance(d, datetime):
                raise ValueError('datetime was not deleted')

        # データを回すだけの簡単なお仕事
        def rec(data: dict):
            for key, val in data.items():
                if isinstance(data[key], list) and _item_literal_check(
                        data[key]):
                    for i in data[key]:
                        detect_date(i)
                elif isinstance(data[key], list):
                    for item in data[key]:
                        detect_ref(item)
                        rec(item)
                elif isinstance(data[key], dict):
                    detect_ref(data[key])
                    rec(data[key])

                # 型変換
                else:
                    detect_date(val)

        actual = self.search.generate_json_dict(data)
        self.assertIsInstance(actual, dict)
        self.assertIsNone(rec(actual))

        # 正常系 refsの指定 resultに_id, ed_parent, _ed_childを残す
        data = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }
        expected = copy.deepcopy(data)
        del expected['coll1']['coll2']['coll3'][0][self.file]
        refs = ['_id', self.parent, self.child]
        actual = self.search.generate_json_dict(data, include=refs)
        self.assertDictEqual(actual, expected)

        # 正常系 refsに設定するが空リストの場合
        data = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }
        expected = copy.deepcopy(data)
        del expected['coll1']['_id']
        del expected['coll1'][self.child]
        del expected['coll1']['coll2']['_id']
        del expected['coll1']['coll2'][self.parent]
        del expected['coll1']['coll2'][self.child]
        del expected['coll1']['coll2']['coll3'][0]['_id']
        del expected['coll1']['coll2']['coll3'][0][self.parent]
        del expected['coll1']['coll2']['coll3'][0][self.child]
        del expected['coll1']['coll2']['coll3'][0][self.file]
        del expected['coll1']['coll2']['coll3'][1]['_id']
        del expected['coll1']['coll2']['coll3'][1][self.parent]
        del expected['coll1']['coll2']['coll3'][1][self.child]
        actual = self.search.generate_json_dict(data, include=[])
        self.assertDictEqual(actual, expected)

        # 異常系 リファレンス系以外のキーを指定した場合
        data_e1 = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }
        with self.assertRaises(ValueError):
            _ = self.search.generate_json_dict(data_e1, include=['test'])

        # 異常系 exclusionにリストとNone以外の値が入力された場合
        data_e2 = {
            'coll1': {
                '_id': ObjectId(),
                'data1': 1,
                self.child: ['aa', 'bb'],
                'coll2': {
                    '_id': ObjectId(),
                    'data2': 2,
                    self.parent: 'cc',
                    self.child: ['dd', 'ee'],
                    'coll3': [
                        {
                            '_id': ObjectId(),
                            'data3': 3,
                            'list_data': [1, 3.24, 'string'],
                            self.parent: 'ff',
                            self.child: ['gg', 'hh'],
                            self.file: [ObjectId(), ObjectId()]
                        },
                        {
                            '_id': ObjectId(),
                            'data4': 4,
                            self.parent: 'ii',
                            self.child: ['jj', 'kk']
                        }
                    ]
                }
            }
        }
        with self.assertRaises(ValueError):
            _ = self.search.generate_json_dict(data_e2, include=(1, 2))

    def test__format_datetime(self):
        # 正常系
        data = ['2018/11/22', '2018/11/22 11:45:23', '2019-01-01 00:00:00']
        for i in map(dateutil.parser.parse, data):
            # 入力値と出力値がdatetime形式として一致しているか確認
            with self.subTest(i=i):
                actual = self.search._format_datetime(i)
                self.assertEqual(i, dateutil.parser.parse(actual[self.date]))

            # 構造のテスト
            self.assertIsInstance(actual, dict)
            self.assertIsInstance(actual[self.date], str)
            self.assertEqual(self.date, list(actual.keys())[0])

        # 正常系 階層指定 中身がdb.find()なので割愛
        # all_docs = self.search.get_documents(1, doc_col, doc_id,
        #                                      parent_depth=0, child_depth=1)
        # print('all_docs:', all_docs)
        # 正常系 単一のドキュメント 中身がdb.find()なので割愛
        # all_docs = self.search.get_documents(3, doc_col, doc_id,
        #                                      parent_depth=0, child_depth=0)
        # print('all_docs:', all_docs)

    def test_get_tree(self):
        if not self.db_server_connect:
            return

        # テストデータ
        parent_col = 'Beamtime'
        target_col = 'expInfo'
        d = {
            parent_col:
                {
                    "test_data": "test",
                    target_col: [
                        {
                            "layer_test1b": "data",
                            "layer_test1a": [
                                {
                                    "test2_data": "data",
                                    "layer_test2": [
                                        {
                                            "layer_test3a": "data",
                                            "layer_test3b": [
                                                {
                                                    "layer_test4a": "data",
                                                    "layer_test4b": [
                                                        {
                                                            "layer_test5a": "data"
                                                        }
                                                    ]
                                                }
                                            ]
                                        }
                                    ]
                                },
                                {
                                    "test3_data": "data",
                                    "layer_test2": [
                                        {
                                            "layer_test3a_2": "data3_2",
                                            "layer_test3b": [
                                                {
                                                    "layer_test4a_2": "data_2",
                                                    "layer_test4b": [
                                                        {
                                                            "layer_test5a_2": "data_2"
                                                        },
                                                        {
                                                            "elen": "elen",
                                                            "layer_test5": [
                                                                {
                                                                    "mario": "mario"
                                                                },
                                                                {
                                                                    "mario2": "mario22"
                                                                },
                                                            ]
                                                        }
                                                    ]
                                                }
                                            ]
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                }
        }

        convert = Convert()
        insert_result = self.db.insert(convert.dict_to_edman(d))

        # インサート結果からtopのoidを取得
        b = []
        for i in insert_result:
            for k, v in i.items():
                if k == parent_col:
                    b.extend(v)
        insert_root_oid = b[0]

        # rootのドキュメントを取得
        docs = self.search.doc2(parent_col, insert_root_oid)

        # tree取得でrootを指定した場合(Config.parentなし)
        top_all_tree = self.search.get_tree(parent_col, docs['_id'])

        # 中間子要素と最後の子要素のテスト用データ取得のために子要素を全部取得
        o = self.db.get_child_all({parent_col: docs})
        x = self.search.generate_json_dict(o, include=['_id', Config.parent,
                                                       Config.child])

        # tree取得で中間の場合
        self_doc_id = x['expInfo'][0]['_id']
        mid_all_tree = self.search.get_tree('expInfo', self_doc_id)

        # tree取得で最後の子要素を指定した場合(Config.childなし)
        self_doc_id = (x['expInfo'][0]['layer_test1a'][1]['layer_test2'][0]
        ['layer_test3b'][0]['layer_test4b'][1]['layer_test5'][0]['_id'])
        last_all_tree = self.search.get_tree('layer_test5', self_doc_id)

        self.assertDictEqual(top_all_tree, mid_all_tree)
        self.assertDictEqual(mid_all_tree, last_all_tree)
        self.assertDictEqual(last_all_tree, top_all_tree)

        # oidを含むツリーを取得する場合(例としてpx-appの詳細画面のtree取得を想定)
        # test_tree = self.search.get_tree('layer_test5', self_doc_id, include=['_id'])
        # print(test_tree)

    def test_doc2(self):
        if not self.db_server_connect:
            return

        # テストデータ
        doc = {
            'test': 'star',
            'val': 456,
            self.parent: ObjectId(),
            self.child: [ObjectId(), ObjectId()],
            self.file: [ObjectId(), ObjectId()]
        }
        collection = 'test_doc'
        insert_result = self.testdb[collection].insert_one(doc)
        oid = insert_result.inserted_id

        # 正常系 リファレンスデータを除く
        actual = self.search.doc2(
            collection, oid,
            exclude_keys=['_id', self.parent, self.child, self.file])
        expected = {'test': 'star', 'val': 456}
        self.assertDictEqual(actual, expected)

        # 正常系 リファレンスデータあり
        actual = self.search.doc2(collection, oid)
        expected = copy.deepcopy(doc)
        self.assertDictEqual(actual, expected)

    # def test_logger_test(self):
    #     self.search.logger_test()

    # def test_find(self):
    #     pass

    # d = {
    #     "Beamtime":
    #         [
    #             {
    #                 "date": {"#date": "2019-09-17"},
    #                 "expInfo": [
    #                     {
    #                         "time": {"#date": "2019/09/17 13:21:45"},
    #                         "int_value": 135,
    #                         "float_value": 24.98
    #                     },
    #                     {
    #                         "time": {"#date": "2019/09/17 13:29:12"},
    #                         "string_value": "hello world"
    #                     },
    #                     {"layer_test1a": {
    #                         "layer_test2": {
    #                             "layer_test3a": "data",
    #                             "layer_test3b": {
    #                                 "layer_test4a": "data",
    #                                 "layer_test4b": {
    #                                     "layer_test5a": "data"
    #                                 },
    #                                 "data": "data"
    #                             },
    #                         },
    #                         "test2_data": "data"
    #                     },
    #                         "layer_test1b": "data"}
    #                 ]
    #             },
    #             {
    #                 "date": {"#date": "2019-09-18"},
    #                 "expInfo": [
    #                     {
    #                         "array_value": ["string", 1234, 56.78, True,
    #                                         None],
    #                         "Bool": False,
    #                         "Null type": None
    #                     }
    #                 ]
    #             }
    #         ]
    # }

    # OKパターン
    # d = {
    #     "a_col": {
    #         "key1": "data"
    #     }
    # }

    # OKパターン
    # d = {
    #     "a_col": {
    #         "b_col": {"key":"data"}
    #     }
    # }

    # # NGパターン
    # d = {
    #     "a_col": {
    #         "b_col": {
    #             "c_col":{"key":"data"}
    #         }
    #     }
    # }

    # # OKパターン
    # d = {
    #     "a_col": {
    #         "key":"data",
    #         "b_col": {
    #             "key":"data",
    #             "c_col":{"key":"data"}
    #         }
    #     }
    # }
    #

    # convert = Convert()
    # converted_edman = convert.dict_to_edman(d)
    # print(f"{converted_edman=}")
    # insert_result = self.db.insert(converted_edman)
    # print(f"{insert_result=}")

    # b= []
    # for i in insert_result:
    #     for k ,v in i.items():
    #         if k == 'Beamtime':
    #             b.extend(v)
    #
    # oid = b[0]
    # r = self.search.find('Beamtime',{'_id':oid},parent_depth=1, child_depth=5)
    # print(f"{r=}")

    # d = DB({'port': '27017', 'host': 'localhost', 'user': 'admin',
    #         'password': 'admin', 'database': 'pxs_edman-user02_db',
    #         'options': ['authSource=admin']})
    # s = Search(d)
    # collection = 'plate'
    # query = {'_id': ObjectId('655c0b0854e5efe89ad26747')}
    # search_result = s.find(collection, query, parent_depth=0,
    #                        child_depth=1, exclusion=['_id'])
    # print('result', search_result)

    # r = d.get_child_all(search_result)
    # r = d.get_child(search_result, 0)
    # print('r', r)

    #
    # def test__self_data_select(self):
    #     # 画面上の選択処理なので、テストは割愛
    #     pass
    #



import os
from datetime import datetime
from enum import Enum, auto
from logging import INFO, getLogger
from pathlib import Path

from bson.json_util import dumps

from edman.exceptions import EdmanFormatError


class JsonManager:
    """
    JSONファイルの取扱いクラス
    """

    def __init__(self) -> None:
        # ログ設定(トップに伝搬し、利用側でログとして取得してもらう)
        self.logger = getLogger(__name__)
        self.logger.setLevel(INFO)
        self.logger.propagate = True

    @staticmethod
    def save(report_data: dict, path: str | Path, name: str,
             date=False) -> None:
        """
        JSONファイルに書き出し

        :param dict report_data: 対象の辞書データ
        :param path: ファイルパス
        :type path: str or Path
        :param str name: ファイル名
        :param bool date: 日付 ファイル名先頭に追加
        :return: None
        """
        if not isinstance(report_data, dict):
            raise EdmanFormatError('Not Dict Data')

        date_str = datetime.today().strftime(
            "%Y%m%d%H%M%S%f") + "_" if date else ""
        filename = date_str + name + '.json'
        p = path if isinstance(path, Path) else Path(path)
        savepath = p / filename

        with savepath.open("w", encoding='utf8') as file:
            file.write(dumps(report_data, ensure_ascii=False, indent=4))
            file.flush()
            os.fsync(file.fileno())


class GetJsonStructure(Enum):
    manual_select = auto()
    all_doc = auto()
    uni_doc = auto()

    @staticmethod
    def members():
        return [*GetJsonStructure.__members__.values()]
